from datetime import datetime
from importlib.metadata import version
from pathlib import Path

from dbt_platform_helper.constants import IMAGE_TAG_ENV_VAR
from dbt_platform_helper.constants import PLATFORM_HELPER_VERSION_OVERRIDE_KEY
from dbt_platform_helper.constants import SERVICE_CONFIG_FILE
from dbt_platform_helper.constants import SERVICE_DIRECTORY
from dbt_platform_helper.constants import (
    TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE_ENV_VAR,
)
from dbt_platform_helper.constants import TERRAFORM_MODULE_SOURCE_TYPE_ENV_VAR
from dbt_platform_helper.domain.terraform_environment import (
    EnvironmentNotFoundException,
)
from dbt_platform_helper.entities.service import ServiceConfig
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.config import ConfigLoader
from dbt_platform_helper.providers.config import ConfigProvider
from dbt_platform_helper.providers.config_validator import ConfigValidator
from dbt_platform_helper.providers.environment_variable import (
    EnvironmentVariableProvider,
)
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.providers.terraform_manifest import TerraformManifestProvider
from dbt_platform_helper.providers.yaml_file import YamlFileProvider
from dbt_platform_helper.utils.application import load_application
from dbt_platform_helper.utils.deep_merge import deep_merge

# TODO add schema version too service config


class ServiceManager:
    def __init__(
        self,
        config_provider=ConfigProvider(ConfigValidator()),
        loader: ConfigLoader = ConfigLoader(),
        io: ClickIOProvider = ClickIOProvider(),
        file_provider=YamlFileProvider,
        environment_variable_provider: EnvironmentVariableProvider = None,
        manifest_provider: TerraformManifestProvider = None,
        platform_helper_version_override: str = None,
        load_application=load_application,
    ):

        self.file_provider = file_provider
        self.config_provider = config_provider
        self.loader = loader
        self.io = io
        self.environment_variable_provider = (
            environment_variable_provider or EnvironmentVariableProvider()
        )
        self.manifest_provider = manifest_provider or TerraformManifestProvider()
        self.platform_helper_version_override = (
            platform_helper_version_override
            or self.environment_variable_provider.get(PLATFORM_HELPER_VERSION_OVERRIDE_KEY)
        )
        self.load_application = load_application

    def generate(self, environments: list[str], services: list[str], image_tag_flag: str = None):

        config = self.config_provider.get_enriched_config()
        application_name = config.get("application", "")
        application = self.load_application(app=application_name)

        if not environments:
            for environment in application.environments:
                environments.append(environment)
        else:
            for environment in environments:
                if environment not in application.environments:
                    raise EnvironmentNotFoundException(
                        f"cannot generate terraform for environment {environment}.  It does not exist in your configuration"
                    )

        if not services:
            try:
                for dir in Path("services").iterdir():
                    if dir.is_dir():
                        config_path = dir / SERVICE_CONFIG_FILE
                        if config_path.exists():
                            services.append(dir.name)
                        else:
                            self.io.warn(
                                f"Failed loading service name from {dir.name}.\n"
                                "Please ensure that your '/services' directory follows the correct structure (i.e. /services/<service_name>/service-config.yml) and the 'service-config.yml' contents are correct."
                            )
            except Exception as e:
                self.io.abort_with_error(f"Failed extracting services with exception, {e}")
        service_models = []
        for service in services:
            service_models.append(
                self.loader.load_into_model(
                    f"{SERVICE_DIRECTORY}/{service}/{SERVICE_CONFIG_FILE}",
                    ServiceConfig,
                )
            )

        platform_helper_version_for_template: str = (
            self.platform_helper_version_override
            or config.get("default_versions", {}).get("platform-helper")
        )

        source_type = self.environment_variable_provider.get(TERRAFORM_MODULE_SOURCE_TYPE_ENV_VAR)

        if source_type == "LOCAL":
            module_source_override = ServiceConfig.local_terraform_source
        elif source_type == "OVERRIDE":
            module_source_override = self.environment_variable_provider.get(
                TERRAFORM_ECS_SERVICE_MODULE_SOURCE_OVERRIDE_ENV_VAR
            )
        else:
            module_source_override = None

        image_tag = image_tag_flag or self.environment_variable_provider.get(IMAGE_TAG_ENV_VAR)
        if not image_tag:
            raise PlatformException(
                f"An image tag must be provided to deploy a service. This can be set by the $IMAGE_TAG environment variable, or the --image-tag flag."
            )

        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        for service in service_models:

            for environment in environments:

                model_dump = service.model_dump(exclude_none=True)
                env_overrides = model_dump.get("environments", {}).get(environment)
                if env_overrides:
                    merged_config = deep_merge(model_dump, env_overrides)
                else:
                    merged_config = model_dump.copy()
                merged_config.pop("environments", None)

                output_path = Path(
                    f"terraform/{SERVICE_DIRECTORY}/{environment}/{service.name}/{SERVICE_CONFIG_FILE}"
                )
                output_path.parent.mkdir(parents=True, exist_ok=True)

                self.file_provider.write(
                    str(output_path),
                    merged_config,
                    f"# WARNING: This is an autogenerated file, not for manual editing.\n# Generated by platform-helper {version('dbt-platform-helper')} / {timestamp}.\n",
                )

                self.manifest_provider.generate_service_config(
                    service,
                    environment,
                    image_tag,
                    platform_helper_version_for_template,
                    config,
                    module_source_override,
                )



from typing import ClassVar
from typing import Dict
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Field


class HealthCheck(BaseModel):
    path: Optional[str] = Field(description="""Path the healthcheck calls""", default=None)
    port: Optional[int] = Field(description="""Port the healthcheck calls""", default=None)
    success_codes: Optional[str] = Field(
        description="""The success codes the healthcheck looks for""", default=None
    )
    healthy_threshold: Optional[int] = Field(description="""The number of  """, default=None)
    unhealthy_threshold: Optional[int] = Field(description="""The number of """, default=None)
    interval: Optional[str] = Field(
        description="""The interval inbetween health check calls""", default=None
    )
    timeout: Optional[str] = Field(
        description="""The timeout for a healthcheck call""", default=None
    )
    grace_period: Optional[str] = Field(description="""The time""", default=None)


class Http(BaseModel):
    path: Optional[str] = Field(
        description="""Requests to this path will be forwarded to your service.""", default=None
    )
    alb: Optional[str] = Field(default=None)
    target_container: Optional[str] = Field(
        description="""Target container for the requests""", default=None
    )
    healthcheck: Optional[HealthCheck] = Field(default=None)


class Sidecar(BaseModel):
    port: int = Field()
    image: str = Field()
    essential: Optional[bool] = Field(default=None)
    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)


class SidecarOverride(BaseModel):
    port: Optional[int] = Field(default=None)
    image: Optional[str] = Field(default=None)
    essential: Optional[bool] = Field(default=None)
    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)


class Image(BaseModel):
    build: Optional[str] = Field(default=None)
    location: Optional[str] = Field(default=None)
    port: Optional[int] = Field(default=None)


class VPC(BaseModel):
    placement: str = Field()


class Network(BaseModel):
    connect: bool = Field(
        description="Enable Service Connect for intra-environment traffic between services."
    )
    vpc: Optional[VPC] = Field(default=None)


class Storage(BaseModel):
    readonly_fs: bool


class ServiceConfigEnvironmentOverride(BaseModel):
    http: Optional[Http] = Field(default=None)
    sidecars: Optional[Dict[str, SidecarOverride]] = Field(default=None)
    image: Optional[Image] = Field(default=None)

    cpu: Optional[int] = Field(default=None)
    memory: Optional[int] = Field(default=None)
    count: Optional[int] = Field(default=None)
    exec: Optional[bool] = Field(default=None)
    network: Optional[Network] = Field(default=None)

    storage: Optional[Storage] = Field(default=None)

    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)


class ServiceConfig(BaseModel):
    name: str = Field(description="""Name of the Service.""")
    type: str = Field(description="""The type of service""")

    http: Optional[Http] = Field(default=None)  # TODO http required if service type load balancer
    sidecars: Optional[Dict[str, Sidecar]] = Field(default=None)
    image: Image = Field()

    cpu: int = Field()
    memory: int = Field()
    count: int = Field()
    exec: Optional[bool] = Field(default=None)
    network: Optional[Network] = Field(default=None)

    storage: Optional[Storage] = Field(default=None)

    variables: Optional[Dict[str, Union[str, int, bool]]] = Field(default=None)
    secrets: Optional[Dict[str, str]] = Field(default=None)
    # Environment overrides can override almost the full config
    environments: Optional[Dict[str, ServiceConfigEnvironmentOverride]] = Field(default=None)

    # Class based variable used when handling the obejct
    local_terraform_source: ClassVar[str] = "../../../../../platform-tools/terraform/ecs-service"



import click

from dbt_platform_helper.domain.service import ServiceManager
from dbt_platform_helper.domain.versioning import PlatformHelperVersioning
from dbt_platform_helper.platform_exception import PlatformException
from dbt_platform_helper.providers.io import ClickIOProvider
from dbt_platform_helper.utils.click import ClickDocOptGroup


@click.group(cls=ClickDocOptGroup)
def service():
    """Commands affecting services."""
    PlatformHelperVersioning().check_if_needs_update()


@service.command(help="Generate terraform manifest for the specified service(s).")
@click.option(
    "--name",
    "-n",
    required=False,
    help="The name of the service to generate a manifest for. Multiple values accepted.",
    multiple=True,
)
@click.option(
    "--environment",
    "-e",
    required=False,
    multiple=True,
    help="The name of the environment to generate service manifests for. Multiple values accepted.",
)
@click.option(
    "--image-tag",
    "-i",
    required=False,
    help="Docker image tag to deploy for the service. Overrides the $IMAGE_TAG environment variable.",
)
def generate(name, environment, image_tag):
    """Validates the service-config.yml format, applies the environment-specific
    overrides, and generates a Terraform manifest at
    /terraform/services/<environment>/<service>/main.tf.json."""

    services = list(name)
    environments = list(environment)
    click_io = ClickIOProvider()

    try:
        service_manager = ServiceManager()
        service_manager.generate(
            environments=environments, services=services, image_tag_flag=image_tag
        )

    except PlatformException as err:
        click_io.abort_with_error(str(err))
